#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
test_atws
----------------------------------

Tests for `atws` module.
"""


import sys
import unittest
from contextlib import contextmanager
from click.testing import CliRunner
import atws
from atws import create_picklist_module

query_test_output=u'<?xml version="1.0" ?>\n<queryxml>\n\t<entity>Ticket</entity>\n\t<query>\n\t\t<field>\n\t\t\tStatus\n\t\t\t<expression op="NotEqual">5</expression>\n\t\t</field>\n\t\t<condition>\n\t\t\t<condition operator="OR">\n\t\t\t\t<field>\n\t\t\t\t\tIssueType\n\t\t\t\t\t<expression op="GreaterThan">345</expression>\n\t\t\t\t</field>\n\t\t\t</condition>\n\t\t</condition>\n\t</query>\n</queryxml>\n'

class TestAtws(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_000_zone_lookup_failure(self):
        try:
            _ = atws.connect(username='failed@toresolve.com',
                              password='notright')
        except ValueError as e:
            assert 'failed@toresolve.com failed to resolve to a zone' in str(e) 
    
    
    def test_001_query_building_output(self):
        query = atws.Query('Ticket')
        query.WHERE('Status', query.NotEqual, 5)
        query.open_bracket()
        query.OR('IssueType', query.GreaterThan, 345)
        query_output = query.pretty_print()
        assert repr(query_test_output) == repr(query_output)
        
        
    def test_command_line_interface(self):
        runner = CliRunner()
        #         result = runner.invoke(create_picklist_module.main)
        #         assert result.exit_code == 0
        #         assert 'atws.create_picklist_module.main' in result.output
        help_result = runner.invoke(create_picklist_module.main, ['--help'])
        assert help_result.exit_code == 0
        assert 'Show this message and exit.' in help_result.output


if __name__ == '__main__':
    sys.exit(unittest.main())



'''
Created on 10 Jan 2016

@author: matt
'''
from __future__ import absolute_import
from . import monkey_patch
from suds.sudsobject import asdict
from ..helpers import has_udfs, get_udfs, del_user_defined_fields_attribute


def convert_entity_to_dict(entity):
    if has_udfs(entity):
        udfs = get_udfs(entity)
        udfs_list = []
        for udf in udfs:
            udfs_list.append(asdict(udf))
        entity.UserDefinedFields = udfs_list
    else:
        del_user_defined_fields_attribute(entity)
    return asdict(entity)


def mp_as_dict(entity):
    return convert_entity_to_dict(entity)


generic_patches = {
   'asdict':mp_as_dict
   }

monkey_patch.add_generic_patches(generic_patches)


'''
Created on 17 Oct 2015

@author: matt
@todo: check if simply patching 'Entity' class is enough for the Entity patches
'''
from future.utils import iteritems
BLACKLISTED_TYPES = [
    'ATWSError',
    'ATWSResponse',
    'ATWSZoneInfo',
    'ArrayOfATWSError',
    'ArrayOfEntity',
    'ArrayOfEntityInfo',
    'ArrayOfEntityReturnInfo',
    'ArrayOfField',
    'ArrayOfPickListValue',
    'ArrayOfUserDefinedField',
    'Entity',
    'EntityDuplicateStatus',
    'EntityInfo',
    'EntityReturnInfo',
    'EntityReturnInfoDatabaseAction',
    'Field',
    'PickListValue',
    'UserDefinedField',
    ]
    
def get_api_types(suds_object,blacklist = []):
    return [sudtype[0].name 
            for sudtype 
            in suds_object.sd[0].types 
            if sudtype[0].name not in blacklist]


def get_api_objects(suds_object,blacklist = []):
    types = get_api_types(suds_object, blacklist) 
    return [suds_object.factory.create(t) for t in types]


def get_api_classes(suds_object,blacklist = []):
    objects = get_api_objects(suds_object, blacklist)
    return [obj.__class__ for obj in objects]


def monkey_patch(obj,patch_name,patch_obj):
    setattr(obj,patch_name,patch_obj)


class DuckPunch(object):

    
    GENERIC_PATCHES = []
    SPECIFIC_PATCHES = []

    def add_generic_patches(self,patches):
        self.GENERIC_PATCHES.append(patches)
        
        
    def add_specific_patches(self,patches):
        self.SPECIFIC_PATCHES.append(patches)

    
    def __call__(self,wrapper):
        self._process_wrapper(wrapper)
        
        
    def _process_wrapper(self,wrapper):
        self._wrapper = wrapper
        self._wrapper._mp = self
        
        entity_classes = get_api_classes(wrapper.client,BLACKLISTED_TYPES)
        for entity_class in entity_classes:
            self._patch_class(entity_class)
            
            
    def _patch_class(self,entity_class):
        entity_class._wrapper = self._wrapper
        self._specific_patch(entity_class)
        self._generic_patch(entity_class)
        
    
    def _generic_patch(self,entity_class):
        entity_class._wrapper = self._wrapper
        for patches in self.GENERIC_PATCHES:
            for attr_name,attr_obj in iteritems(patches):
                monkey_patch(entity_class,attr_name,attr_obj)
        
    
    def _specific_patch(self,entity_class):
        for patches in self.SPECIFIC_PATCHES:
            for attr_name,attr_obj in patches.get(entity_class.__name__,{}):
                monkey_patch(entity_class,attr_name,attr_obj)
            
    



'''
Created on 17 Oct 2015

@author: matt
'''
from __future__ import absolute_import
from . import monkey_patch
from ..helpers import get_udf_value, set_udf
import json


ATTRIBUTE_UDF_NAME = 'AutomatedSystemTags'


def encode(d):
    return json.dumps(d)


def decode(s):
    return json.loads(s)


def del_attribute(wrapper,entity,name):
    #: :type attributes: dict
    attributes = get_attributes(wrapper, entity)
    del attributes[name]
    set_attributes(wrapper, entity, attributes)
    

def get_attribute(wrapper,entity,name,default=[]):
    attributes = get_attributes(wrapper,entity)
    try:
        result = attributes[name]
    except KeyError:
        if default == []:
            raise
        result = set_attribute(wrapper,entity,name,default)
    return result


def get_attributes(wrapper,entity):
    udf_value = get_udf_value(wrapper,entity,ATTRIBUTE_UDF_NAME,json.dumps({}))
    if not udf_value.strip():
        return {}
    return decode(udf_value)


def set_attribute(wrapper,entity,name,value):
    try:
        value = value.isoformat()
    except (KeyError,TypeError,AttributeError):
        pass
    attributes = get_attributes(wrapper,entity)
    attributes[name] = value
    set_attributes(wrapper,entity,attributes)
    return value


def set_attributes(wrapper,entity,attributes):
    udf_value = encode(attributes)
    set_udf(wrapper,entity,ATTRIBUTE_UDF_NAME,udf_value)


def mp_del_attribute(entity,name):
    return del_attribute(entity._wrapper, entity, name)


def mp_set_attribute(entity,name,value):
    return set_attribute(entity._wrapper, entity, name, value)


def mp_set_attributes(entity,attributes):
    return set_attributes(entity._wrapper, entity, attributes)


def mp_get_attribute(entity,name,default=[]):
    return get_attribute(entity._wrapper, entity, name, default)


def mp_get_attributes(entity):
    return get_attributes(entity._wrapper, entity)

generic_patches = {
   'get_attribute':mp_get_attribute,
   'get_attributes':mp_get_attributes,
   'set_attribute':mp_set_attribute,
   'set_attributes':mp_set_attributes,
   'del_attribute':mp_del_attribute
   }

monkey_patch.add_generic_patches(generic_patches)


from __future__ import absolute_import
import sys
from datetime import datetime
from xml.etree.ElementTree import Element, SubElement, tostring
from xml.sax.saxutils import escape
from .helpers import datetime_to_api_timezone
from .constants import (AUTOTASK_API_QUERY_ID_LIMIT,
                        AUTOTASK_API_QUERY_DATEFORMAT,
                        WRAPPER_DEFAULT_GET_ALL_ENTITIES)

PY3 = sys.version_info >= (3, 0)
QUERY_ENCODING = None
if PY3:
    QUERY_ENCODING = 'unicode'


XML_QUERY_ESCAPE = { '"' : '&quot;',
                     "'" : '&apos;'}

def get_userdefined_field_list_items(wrapper,entity):
    query = Query('UserDefinedFieldListItem')
    query.WHERE('UdfFieldId', query.Equals, entity.id)
    return wrapper.query(query).fetch_all()


def get_entity_by_id(wrapper,entity_type,entity_id):
    result = get_entities_by_field_equals(wrapper,
                                          entity_type,
                                          'id',
                                          entity_id,
                                          False)
    return result[0]


def get_entities_by_field_equals(wrapper,entity_type,field,value,udf=False):
    query = Query(entity_type)
    query.WHERE(field,query.Equals,value,udf)
    return wrapper.query(query) 


def format_datetime_for_api_query(dt):
    return datetime_to_api_timezone(dt).strftime(AUTOTASK_API_QUERY_DATEFORMAT)


def query_escape(value):
    return escape(value, XML_QUERY_ESCAPE)

    
def get_id_query(entity_type,id_list):
    query = Query(entity_type)
    for entity_id in id_list:
        query.OR('id',query.Equals,entity_id)
    return query
    

def get_queries_for_entities_by_id(entity_type,
                                   id_list,
                                   id_limit=AUTOTASK_API_QUERY_ID_LIMIT,
                                   query_function=get_id_query):
    queries = yield_queries_for_entities_by_id(entity_type, 
                                               id_list, 
                                               id_limit, 
                                               query_function)
    return list(queries)


def yield_queries_for_entities_by_id(entity_type,
                                   id_list,
                                   id_limit=AUTOTASK_API_QUERY_ID_LIMIT,
                                   query_function=get_id_query):
    query_ids = []
    for _id in id_list:
        if len(query_ids) == id_limit:
            yield query_function(entity_type, query_ids)
            query_ids = []
        else:
            query_ids.append(_id)
    if query_ids:
        yield query_function(entity_type, query_ids)


class Query(object):
    Equals='Equals'
    NotEqual='NotEqual'
    GreaterThan='GreaterThan'
    LessThan='LessThan'
    GreaterThanorEquals='GreaterThanorEquals'
    LessThanOrEquals='LessThanOrEquals'
    BeginsWith='BeginsWith'
    EndsWith='EndsWith'
    Contains='Contains'
    IsNotNull='IsNotNull'
    IsNull='IsNull'
    IsThisDay='IsThisDay'
    Like='Like'
    NotLike='NotLike'
    SoundsLike='SoundsLike'
    
    get_all_entities = None


    def FROM(self,entity_type):
        self.entity_type = entity_type

        
    def WHERE(self,field_name,field_condition,field_value,udf=False):
        self._add_field(None, field_name, field_condition, field_value, udf)

    
    def OR(self,field_name,field_condition,field_value,udf=False):
        self._add_field('OR', field_name, field_condition, field_value, udf)
    
    
    def AND(self,field_name,field_condition,field_value,udf=False):
        self._add_field('AND', field_name, field_condition, field_value, udf)
        
        
    def open_bracket(self,operator=None):
        attrib = {}
        if operator:
            attrib = {'operator':operator}
        self._cursor = SubElement(self._cursor,'condition',attrib=attrib)
    
    
    def close_bracket(self):
        self._close_cursor()
        

    def reset(self):
        self._query_elements = []
        self._query_elements.append(self._query)
        self._query.clear()
        self.minimum_id_xml = None
        self.minimum_id = None
        self.minimum_id_field = 'id'

    
    def get_query_xml(self):
        self._entityxml.text = self.entity_type
        if self.minimum_id:
            self._add_min_id_field()
        query_xml = tostring(self._queryxml, encoding=QUERY_ENCODING)
        #query_xml = query_xml.encode('utf-8')
        return query_xml
    
    
    def pretty_print(self):
        import xml.dom.minidom
        return xml.dom.minidom.parseString(self.get_query_xml()).toprettyxml()
    
    
    def set_minimum_id(self,minimum_id,field='id'):
        self.minimum_id = minimum_id
        self.minimum_id_field = field
        

    @property
    def _cursor(self):
        return self._query_elements[-1]
    
    
    @_cursor.setter
    def _cursor(self, element):
        self._query_elements.append(element)
        
        
    def _close_cursor(self):
        del(self._query_elements[-1])
        

    def _add_field(self,operator,field_name,field_condition,field_value,udf=False):
        attributes = {}
        if udf:
            attributes['udf'] = 'true'
        if operator:
            self.open_bracket(operator)
        field = SubElement(self._cursor,'field', attrib=attributes)
        field.text = field_name
        expression = SubElement(field,'expression',attrib={'op':field_condition})
        expression.text = self._process_field_value(field_value)
        if operator:
            self.close_bracket()
        return field,expression
                    
        
    def _add_min_id_field(self):
        try:
            self._update_min_id_xml()
        except AttributeError:
            self._create_min_id_xml()
            
    
    def _update_min_id_xml(self):
        self.minimum_id_xml.text = self._process_field_value(self.minimum_id)
    
    
    def _create_min_id_xml(self):
        minimum_id = self._process_field_value(self.minimum_id)
        # use _add_field instead of self.AND because we need to keep the
        # expression
        expression = self._add_field('AND',
                                     self.minimum_id_field,
                                     self.GreaterThan,
                                     minimum_id)[1]
        self.minimum_id_xml = expression
    
    
    def _process_field_value(self,value):
        if type(value) is datetime:
            return format_datetime_for_api_query(value)
        return query_escape(str(value))
    
    
    def __init__(self,entity_type = None):
        self.get_all_entities = WRAPPER_DEFAULT_GET_ALL_ENTITIES
        self.entity_type = entity_type
        self._queryxml = Element('queryxml')
        self._entityxml = SubElement(self._queryxml, 'entity')
        self._query = SubElement(self._queryxml, 'query')
        self.reset()


    def __str__(self):
        return repr(self.get_query_xml())
    


# -*- coding: utf-8 -*-
from __future__ import absolute_import
import os
import logging
import click
from .wrapper import connect
from .helpers import create_atvar_module
from .monkeypatch.duckpunch import BLACKLISTED_TYPES, get_api_types

@click.command()
@click.argument('target_path')
@click.option('--username', 
              default=os.environ.get('AUTOTASK_API_USERNAME', None), 
              help='your Autotask username')
@click.option('--password',
              default=os.environ.get('AUTOTASK_API_PASSWORD', None), 
              help='your Autotask password')
@click.option('--url',
              default=os.environ.get('AUTOTASK_API_URL', None), 
              help='your Autotask webservices URL (wsdl file)')
@click.option('--entities',
              default=list(), 
              help='Space delimited entity types for inclusion')
@click.option('--log_level',
              default=os.environ.get('LOG_LEVEL', None),
              help='python logging root logger level with basic config')
def main(target_path, username, password, url, entities, log_level):
    if log_level:
        logging.basicConfig(level=getattr(logging,log_level))
    at = connect(url=url, username=username, password=password)
    if not entities:
        entities = get_api_types(at.client, BLACKLISTED_TYPES + ['Currency'])
    
    create_atvar_module(at, entities, target_path)


if __name__ == "__main__":
    main()


'''
Created on 17 Oct 2015

@author: matt
'''
from __future__ import absolute_import
from suds.sudsobject import Object as sudsobject
from . import monkey_patch
from .. import helpers

def mp_set_udf(entity, name, value):
    helpers.set_udf(entity._wrapper, entity, name, value)
    

def mp_get_udf_value(entity,name):
    return helpers.get_udf_value(entity._wrapper, entity, name)


def mp_del_udf(entity,name):
    helpers.del_udf(entity._wrapper, entity, name)
    

def mp_getattr(entity,attr):
    if attr != 'UserDefinedFields':
        try:
            return helpers.get_udf_value(entity._wrapper,entity,attr)
        except AttributeError:
            raise AttributeError( 'no attribute or udf named {}'.format(attr) )
    return sudsobject.__getattribute__(entity,attr)


def mp_setattr(self,name,value):
    try:
        udf = helpers.get_udf(self._wrapper,self,name)
    except AttributeError:
        sudsobject.__setattr__(self,name,value)
    else:
        udf.Value = value
        

generic_patches = {
   'set_udf':mp_set_udf,
   'get_udf':mp_get_udf_value,
   'del_udf':mp_del_udf,
   '__getattr__':mp_getattr,
   '__setattr__':mp_setattr
   }
monkey_patch.add_generic_patches(generic_patches)



'''
Created on 10 Jan 2016

@author: matt
'''
from __future__ import absolute_import
from future.utils import iteritems
from builtins import str
import logging
import pytz
from decimal import Decimal
from datetime import datetime
from past.builtins import basestring
from . import monkey_patch
from . import asdict
from ..helpers import localise_datetime

logger = logging.getLogger(__name__)
UTC = pytz.timezone('UTC')
MARSHAL_MAP = {}


def datetime_to_utc_isoformat(dt):
    try:
        return localise_datetime(dt).astimezone(UTC).isoformat()
    except OverflowError:
        return dt.isoformat()
    except Exception:
        logger.exception('failed to isoformat dt: %s', dt)
        raise


def convert(obj):
    result = {}
    for k,v in iteritems(obj):
        if type(v) == list:
            result[k] = []
            for d in v:
                result[k].append(convert(d))
        else:
            result[k] = convert_value(v)
    logger.debug('returning the following dictionary as marshallable')
    logger.debug(result)
    return result


def convert_value(v):
    if isinstance(v, datetime):
        try:
            return datetime_to_utc_isoformat(v)
        except AttributeError:
            return v
    if isinstance(v, basestring):
        try:
            return str(v)
        except TypeError:
            return v
    if isinstance(v, Decimal):
        return str(v)
    return v


def get_marshallable_dict(entity):
    e = entity.asdict()
    return convert(e)


generic_patches = {
   'asmarshallabledict':get_marshallable_dict,
   'asserializabledict':get_marshallable_dict
   }

monkey_patch.add_generic_patches(generic_patches)


# -*- coding: utf-8 -*-
from __future__ import absolute_import
from pytz import timezone
from .wrapper import connect
from .query import Query

__author__ = 'Matt Parr'
__email__ = 'matt@parr.geek.nz'
__version__ = '0.4.8'

ATWS_API_TIMEZONE = timezone('US/Eastern')
ATWS_ENTITY_SEND_LIMIT = 200
ATWS_ENTITY_RECEIVE_LIMIT = 500



'''
Created on 17 Oct 2015

@author: matt
'''
from __future__ import absolute_import
from .. import query
from ..helpers import get_entity_type,copy_attributes
from . import monkey_patch


def mp_update(entity):
    result = entity._wrapper.update(entity).fetch_one()
    copy_attributes(result,entity)
    return True


def mp_delete(entity):
    entity._wrapper.delete(entity).execute()
    return True


def mp_reload(entity):
    query = query.Query(get_entity_type(entity))
    query.WHERE('id',query.Equals,entity.id)
    result = entity._wrapper.query(query).fetch_one()
    copy_attributes(result,entity)
    return True
    

def mp_create(entity):
    result = entity._wrapper.create(entity).fetch_one()
    copy_attributes(result,entity)
    return True
    
generic_patches = {
                   'create':mp_create,
                   'reload':mp_reload,
                   'update':mp_update,
                   'delete':mp_delete
                   }
monkey_patch.add_generic_patches(generic_patches)



'''
Created on 3 Nov 2015

@author: matt

@todo: monkeypunch lazy picklists into entities by default combined with below
@todo: create a switch to turn on automatic picklist value reversal
@todo: update create_picklist_module.py to use this and child naming 
'''
from __future__ import absolute_import
import os
from cached_property import cached_property
from .helpers import get_field_info


def is_active(*args):
    return args[0]['IsActive']
    
    
def always_true(*args):
    return True


def find(index_label, index_name, object_list, condition = always_true):
    try:
        index = next(index for (index, d) in enumerate(object_list) 
                    if str(d[index_label]) == str(index_name) and condition(d))
    except StopIteration:
        raise KeyError('label not found in index', index_label, index_name)
    else:
        return object_list[index]
    

def get_field_picklist(field_name, field_info):
    field_picklist = find('Name', 
                          field_name, 
                          field_info.Field)
    return field_picklist


def is_child_field(field_picklist):
    child_parent_field_name = get_child_parent_field_name(field_picklist) 
    return child_parent_field_name is not None


def get_child_parent_field_name(field_picklist):
    parent_field_name = field_picklist.PicklistParentValueField
    return parent_field_name


def get_label_value(label, picklistvalues, condition=is_active):
    item = find('Label', label, picklistvalues, condition=condition)
    return item.Value


def get_value_label(value, picklistvalues, condition=is_active):
    item = find('Value', value, picklistvalues, condition=condition)
    return item.Label


def get_child_label_value(label, picklistvalues, condition):
    item = find('Label', label, picklistvalues, condition = condition)
    return item.Value


def child_picklist_as_dict(child_picklist):
    child_field_dict = {}
    parent = child_picklist.parent
    for item in child_picklist._picklist:
        parent_item_label = parent.reverse_lookup(item.parentValue,
                                                  condition=always_true)
        parent_item_dict = child_field_dict.setdefault(parent_item_label, {})
        parent_item_dict[item.Label] = item.Value
        
    return {child_picklist.entity_type:
            {child_picklist.field_name:child_field_dict}}
    
    
def picklist_as_dict(picklist):
    return {picklist.entity_type:
            {picklist.field_name:
             {item.Label:item.Value for item in picklist._picklist}}}
    
    
class ChildFieldPicklist(object):
    def __init__(self, parent, parent_item_label, field_name):
        self.field_name = field_name
        self._parent = parent
        self.parent_item_label = parent_item_label
        
    
    @property
    def __name__(self):
        return self.child_picklist_name + '_' + self.parent_label
    
    
    @property
    def parent_item_value(self):
        return self._parent[self.parent_item_label]
    

    @property
    def _picklist(self):
        return self._picklist_info.PicklistValues.PickListValue
    
    
    @property
    def _picklist_info(self):
        return get_field_picklist(self.field_name, 
                                  self._parent.entity_picklists._field_info)
    
        
    def _condition(self, picklist_value):
        return (picklist_value.parentValue == self.parent_item_value 
                and is_active(picklist_value))
    
    
    def lookup(self, label):
        return get_child_label_value(label, self._picklist, self._condition)
    
    
    def __getitem__(self, item):
        return self.lookup(item)
    
    
    def __call__(self, label):
        return self.lookup(label)
    

class FieldPicklist(object):
    def __init__(self, entity_picklists, field_name):
        self.field_name = field_name
        self.entity_picklists = entity_picklists
        if self.is_child:
            self._children = {}
        
        
    @property
    def entity_type(self):
        return self.entity_picklists.entity_type
    
        
    @property
    def parent(self):
        if self.is_child:
            return self.entity_picklists[self.parent_name]
        else:
            raise AttributeError('field does not have a parent')
        
        
    @property
    def parent_name(self):
        if self.is_child:
            return get_child_parent_field_name(self._picklist_info)
        else:
            raise AttributeError('field does not have a parent')
        
        
    @property
    def __name__(self):
        name = self.entity_type + '_' + self.field_name
        return name
    
        
    @property
    def _picklist(self):
        return self._picklist_info.PicklistValues.PickListValue
    
    
    @property
    def _picklist_info(self):
        return get_field_picklist(self.field_name, 
                                  self._field_info)
        
                
    @property
    def _field_info(self):
        return self.entity_picklists._field_info
    
    
    @property
    def is_child(self):
        return is_child_field(self._picklist_info)


    def lookup(self, label):
        ''' take a field_name_label and return the id'''
        if self.is_child:
            try:
                return self._children[label]
            except KeyError:
                self._children[label] = ChildFieldPicklist(self.parent, 
                                                           label, 
                                                           self.field_name) 
                return self._children[label]
        else:
            return get_label_value(label, self._picklist)

    
    def reverse_lookup(self, value, condition=is_active):
        ''' take a field_name_id and return the label '''
        label = get_value_label(value, self._picklist, condition=condition)
        return label 


    def as_dict(self):
        if self.is_child:
            return child_picklist_as_dict(self)
        else:
            return picklist_as_dict(self)
        
        
    def __call__(self, label):
        return self.lookup(label)
        
    
    def __getitem__(self,item):
        return self.lookup(item)
    
    
    def __str__(self):
        return os.linesep.join(
            [ '{}.{}.{} = {}'.format(self.entity_type,
                                     self.field_name,
                                     field.Label,
                                     field.Value)
            for field in self._picklist]
                               )
        
    
    def __iter__(self):
        return iter([field for field in self._picklist])
    
        
class EntityPicklists(object):
    def __init__(self,at, entity_type):
        self._at = at
        self.entity_type = entity_type
        self._entity_fields = {}
    
    
    @property
    def __name__(self):
        return self.entity_type
    
        
    @cached_property
    def _field_info(self):
        result = get_field_info(self._at, self.entity_type) 
        return result
        
        
    @cached_property
    def fields(self):
        return [field for field in self._field_info.Field]
            
    
    @cached_property            
    def picklist_fields(self):
        return [field for field in self.fields if 
                field.IsPickList]
        
        
    def refresh(self,picklists=None):
        del self._field_info
        
            
    def lookup(self,field):
        try:
            return self._entity_fields[field]
        except KeyError:
            self._entity_fields[field] = FieldPicklist(self, field)
            return self._entity_fields[field]
            
        
    def __getitem__(self, item):
        return self.lookup(item)
    
    
    def __call__(self, field):
        return self.lookup(field)
    
    
    def __iter__(self):
        return iter([self.lookup(field.Name) for field in self.picklist_fields])
        
        
class Picklists(object):
    def __init__(self,at):
        self._at = at
        self._entity_picklists = {}
    
    
    def refresh(self,entity_type):
        self._entity_picklists[entity_type].refresh()
            
    
    def lookup(self,entity):
        try:
            entity_picklists = self._entity_picklists[entity]
        except KeyError:
            entity_picklists = EntityPicklists(self._at, entity)
            self._entity_picklists[entity] = entity_picklists 
        return entity_picklists
            
            
    def __getitem__(self, item):
        return self.lookup(item)
    
    
    def __call__(self, entity):
        return self.lookup(entity)
    
    
    def __iter__(self):
        return iter(self._entity_picklists)
        
        


'''
Created on 27 Sep 2015

@author: matt
'''
from __future__ import absolute_import
from future.utils import iteritems
import logging
import math
from suds import WebFault
from .constants import (AUTOTASK_API_QUERY_RESULT_LIMIT,
                        AUTOTASK_API_TIMEZONE,
                        LOCAL_TIMEZONE)

logger = logging.getLogger(__name__)

def copy_attributes(from_entity,to_entity):
    attributes = [field[0] for field in from_entity]
    for attribute in attributes:
        setattr(to_entity,attribute,getattr(from_entity,attribute))


def has_udfs(entity):
    try:
        return len(entity.UserDefinedFields.UserDefinedField) >= 1
    except Exception:
        return False


def del_udf(wrapper,entity,name):
    udf = get_udf(wrapper, entity, name)
    entity.UserDefinedFields.UserDefinedField.remove(udf)
    

def get_entity_type(entity):
    return entity.__class__.__name__


def get_udf_value(wrapper,entity,name,default=[]):
    udf = get_udf(wrapper,entity,name,default)
    try:
        return udf.Value
    except AttributeError:
        if default == []:
            raise
        udf.Value = default
    return udf.Value


def get_udfs(entity):
    return entity.UserDefinedFields.UserDefinedField


def get_api_threshold(wrapper):
    result = wrapper.getThresholdAndUsageInfo()
    message = result.EntityReturnInfoResults.EntityReturnInfo[0].Message.split()
    threshold = int(str(message[1]).strip(";"))
    timeframe = int(str(message[3]).strip(";"))
    requests = int(str(message[5]).strip(";"))
    return threshold,timeframe,requests
    
    
def get_api_usage(wrapper):
    threshold,_,requests = get_api_threshold(wrapper)
    return math.ceil( ( float(requests)/float(threshold) )*100 )


def get_udf(wrapper,entity,name,default=[]):
    if has_udfs(entity):
        for udf in get_udfs(entity):
            if name == udf.Name:
                return udf
    if default == []:
        raise AttributeError('no udf named {}'.format(name))
    udf = wrapper.new('UserDefinedField')
    udf.Name = name
    udf.Value = default
    try:
        udf_list = entity.UserDefinedFields.UserDefinedField
    except AttributeError:
        array_of_udfs = wrapper.new('ArrayOfUserDefinedField')
        udf_list = array_of_udfs.UserDefinedField
        entity.UserDefinedFields = array_of_udfs
    udf_list.append(udf)
    return udf


def set_udf(wrapper,entity,name,value):
    udf = get_udf(wrapper,entity, name, value)
    udf.Value = value


def del_user_defined_fields_attribute(entity):
    if hasattr(entity,'UserDefinedFields'):
        delattr(entity,'UserDefinedFields')    


def can_multiupdate_entities(entities):
    True


def entities_have_userdefined_fields(entities):
    for entity in entities:
        if has_udfs(entity):
            return True
    return False


def entities_have_userdefined_fields_with_values(entities):
    for entity in entities:
        if has_udfs(entity):
            for udf in get_udfs(entity):
                value = getattr(udf,'Value',[])
                if value != []:
                    return True
    return False    


def clean_entities(entities):
    for entity in entities:
        clean_entity(entity)
        yield entity


def query_requires_another_call(result,query):
    if not query.get_all_entities:
        return False
    try:
        query.get_query_xml()
    except AttributeError:
        return False
    if query_result_count(result) == AUTOTASK_API_QUERY_RESULT_LIMIT:
        return True
    return False


def get_result_entities(result):
    try:
        return result.EntityResults.Entity
    except Exception:
        return []
    

def get_highest_id(result,field):
    return max(getattr(entity,field) for entity in get_result_entities(result))


def query_result_count(result):
    return len(get_result_entities(result))


def datetime_to_api_timezone(dt):
    if dt.tzinfo is None:
        api_dt = LOCAL_TIMEZONE.localize(dt).astimezone(AUTOTASK_API_TIMEZONE).replace(tzinfo=None)
    else:
        api_dt = dt.astimezone(AUTOTASK_API_TIMEZONE).replace(tzinfo=None)
    return api_dt


def datetime_to_local_timezone(dt):
    if dt.tzinfo is None:
        api_dt = AUTOTASK_API_TIMEZONE.localize(dt).astimezone(LOCAL_TIMEZONE)
    else:
        api_dt = dt.astimezone(LOCAL_TIMEZONE)
    return api_dt


def localise_datetime(dt):
    if dt.tzinfo is None:
        return LOCAL_TIMEZONE.localize(dt)
    else:
        return dt.astimezone(LOCAL_TIMEZONE)


def split_list_into_chunks(list_to_split,chunk_length):
    items = []
    for item in list_to_split:
        if len(items) == chunk_length:
            yield list
            items = []
        items.append(item)
    if items:
        yield items


def clean_udfs(entity):
    new_udf_list = []
    try:
        for udf in entity.UserDefinedFields.UserDefinedField:
            if hasattr(udf,"Value"):
                new_udf_list.append(udf)
        if new_udf_list:
            entity.UserDefinedFields.UserDefinedField = new_udf_list
    except AttributeError:
        pass
    if not new_udf_list:
        del_user_defined_fields_attribute(entity)


def clean_fields(entity):
    remove = list()
    for field_name,field_value in entity:
        if field_name == 'Fields':
            remove.append(field_name)
            continue
        if field_name != 'UserDefinedFields':
            if field_value in ["",None]:
                remove.append(field_name)
    for field in remove:
        delattr(entity,field)


def clean_entity(entity):
    clean_fields(entity)
    clean_udfs(entity)


def process_fields(entity,functions):
    for (field_name,field_value) in entity:
        if field_name in ('Fields','UserDefinedFields'):
            continue
        for fn in functions:
            fn(entity=entity,name=field_name,value=field_value)


def process_udfs(entity,functions):
    try:
        for udf in entity.UserDefinedFields.UserDefinedField:
            for fn in functions:
                fn(entity=entity,udf=udf)
    except AttributeError:
        pass


def process_entity(entity, functions):
    process_fields(entity, functions)
    process_udfs(entity, functions)


def create_userdefined_field_list_items(wrapper,entity,items):
    list_items = [wrapper.new('UserDefinedFieldListItem',**item) 
                  for item in items]
    return wrapper.create(list_items).execute()


def picklist_stream_formatter(s):
    import re
    # Remove invalid characters
    return re.sub('[^0-9a-zA-Z_]', '', s)


def get_field_info(wrapper,entity_type):
    'type: wrapper :atws.Wrapper'
    fields = wrapper.new('GetFieldInfo')
    fields.psObjectType = entity_type
    return wrapper.GetFieldInfo(fields)


def has_picklist_values(field):
    if field.IsPickList:
        if field.PicklistValues:
            return True
    return False
        

def get_field_picklist(picklist_values):
    return {obj.Label:obj.Value for obj in picklist_values}


def get_picklists(get_field_info_response):
    return {field.Name:get_field_picklist(field.PicklistValues.PickListValue) 
     for field in get_field_info_response.Field 
     if has_picklist_values(field)}
    

def get_picklist_stream(entity_type,picklists):
    import re
    csnregex = re.compile('^\d+$')
    failures = []
    for picklist_name,picklist in iteritems(picklists):
        for field_name,field_value in iteritems(picklist):
            try:
                digits = csnregex.findall(field_value)
                if not digits:
                    field_value = '{}'.format(field_value)
                yield "{}_{}_{} = {}\n".format(
                    entity_type,
                    picklist_stream_formatter(picklist_name),
                    picklist_stream_formatter(field_name),
                    repr(field_value)
                    )
            except Exception as e:
                logger.error('failed to add: %s:%s:%s:%s', 
                             entity_type,
                             picklist_name,
                             field_name, 
                             field_value)
                logger.exception('failed processing a field into the picklist')
                failures.append(e)
    if failures:
        raise Exception(failures)


def create_atvar_module(at,entities,file_name):
    with open(file_name,"w") as a: 
        for entity_type in entities:
            try:
                field_info_response = get_field_info(at,entity_type)
            except WebFault:
                pass
            else:
                picklists = get_picklists(field_info_response)
                for line in get_picklist_stream(entity_type,picklists):
                    a.write( line ) 
                    

def datetime_to_api_timezone_entity(**kwargs):
    try:
        datetime_to_api_timezone(kwargs['udf'].Value)
    except (AttributeError,KeyError,TypeError,OverflowError):
        pass
    else:
        return
    try:
        setattr(kwargs['entity'],
                kwargs['name'],
                datetime_to_api_timezone(kwargs['value']))
    except (AttributeError,KeyError,TypeError,OverflowError):
        pass


def datetime_to_local_timezone_entity(**kwargs):
    try:
        datetime_to_local_timezone(kwargs['udf'].Value)
    except (AttributeError,KeyError,TypeError,OverflowError):
        pass
    else:
        return
    try:
        setattr(kwargs['entity'],
                kwargs['name'],
                datetime_to_local_timezone(kwargs['value']))
    except (AttributeError,KeyError,TypeError,OverflowError):
        pass


def trim_empty_strings_entity(**kwargs):
    try:
        kwargs['udf'].Value = kwargs['udf'].Value.trim()
    except (AttributeError,KeyError,TypeError):
        pass
    else:
        return
    try:
        setattr(kwargs['entity'],
                kwargs['name'],
                kwargs['value'].trim())
    except (AttributeError,KeyError,TypeError):
        pass
    
    
def trim_single_space_strings_entity(**kwargs):
    try:
        if kwargs['udf'].Value.isspace():
            kwargs['udf'].Value = ''
    except (AttributeError,KeyError,TypeError):
        pass
    else:
        return
    try:
        if getattr(kwargs['entity'], kwargs['name']).isspace():
            setattr(kwargs['entity'],
                    kwargs['name'],
                    '')
    except (AttributeError,KeyError,TypeError):
        pass


'''
Created on 27 Sep 2015

@author: matt

'''
from __future__ import absolute_import
from future.utils import iteritems
import os
import logging
import re
import uuid
from .constants import (SUPPORT_FILES_ENABLED,
                        SUPPORT_FILES_LOCATION,
                        MONKEY_PATCHING_ENABLED,
                        WRAPPER_DISABLE_CLEAN_ENTITIES,
                        AUTOTASK_API_ENTITY_SEND_LIMIT)
from .helpers import (get_result_entities,
                      datetime_to_api_timezone_entity,
                      trim_empty_strings_entity,
                      datetime_to_local_timezone_entity,
                      get_field_info,
                      clean_entity,
                      process_entity,
                      query_requires_another_call,
                      get_highest_id,
                      get_entity_type,
                      trim_single_space_strings_entity)
from . import monkeypatch
from .monkeypatch import crud
from .monkeypatch import userdefinedfields
from . import connection
from .picklist import Picklists
from suds import sudsobject, MethodNotFound
from suds.plugin import MessagePlugin


logger = logging.getLogger(__name__)


def connect(**kwargs):
    if SUPPORT_FILES_ENABLED:
        plugin = SupportFilesPlugin()
        client_options = kwargs.setdefault('client_options', {})
        plugins = client_options.setdefault('plugins', [])
        plugins.append(plugin)
    wrapper = connection.connect(atws_version=Wrapper,**kwargs)
    if MONKEY_PATCHING_ENABLED:
        monkeypatch.monkey_patch(wrapper)

    return wrapper 


def enable_support_files(path = None):
    SUPPORT_FILES_ENABLED = True
    if path:
        SUPPORT_FILES_LOCATION = path
    

class SupportFilesPlugin(MessagePlugin):
    '''
    creates files for sending to autotask
    set the variables:
    constants.SUPPORT_FILES_ENABLED = True
    constants.SUPPORT_FILES_LOCATION = 'path to save files'
    '''
    def sending(self, context):
        self.write_file(context.envelope, 'sent')

            
    def received(self, context):
        self.write_file(context.reply, 'received')
    
    
    def write_file(self, output, direction):
        file_name = self.file_name
        with open(file_name,"w") as f:
            logger.info('writing %s xml to %s.xml', 
                        direction,
                        file_name)
            f.write(str(output))
            
    
    @property
    def file_name(self):
        return os.path.join(SUPPORT_FILES_LOCATION, 
                            str(uuid.uuid4()) + '.xml')


class AutotaskAPIException(Exception):
    message = """Autotask API has returned an error.
    Please process the response attribute in this exception to detirmine 
    the extent of this failure.
    """ 
    
    def __init__(self, response):
        self.response = response

    def __str__(self):
        return repr(self.message) + repr(self.response.errors)          


class AutotaskProcessException(Exception):
    message = """During send or receive with the API, an error has 
    occurred.  Please see the exception attribute for details about the error.
    Please see the response attribute as some API commands may have sucessfully 
    completed before the exception occurred
    """
    
    def __init__(self, exception, response):
        self.exception = exception
        self.response = response
    
    
    def __str__(self):
        return repr(self.message) + repr(self.exception) + repr(self.response.errors)
    
    
class Response(object):
    
    def add_error(self,error):
        logger.error(error)
        self.errors.append(error)


    def add_entities(self,entities):
        self.response_count.append(len(entities))
        self._wrapper.process_inbound_entities(entities)
        return entities
            

    def raise_or_return_entities(self):
        if self.errors:
            self._raise()
    
    
    def _raise(self):
        raise AutotaskAPIException(self)
    
    
    def __init__(self,wrapper):
        self._wrapper = wrapper
        self.response_count = []
        self.errors = []
        self.successful_entities = []
        self.failed_entities = []    

        
    def _get_errors(self,result):
        try:
            return result.Errors.ATWSError
        except Exception:
            return []


class ResponseAction(Response):
    
    def add_result(self,result,packet):
        if result.ReturnCode == 1:
            return self._record_successful_entities(result)
        else:
            return self._record_failure(result,packet.Entity)


    def _record_successful_entities(self,result):
        entities = get_result_entities(result)
        return self.add_entities(entities)        

    
    def _record_failure(self,result,entities):
        successful_entities = self._record_successful_entities(result)
        errors = [self._get_error(error,entities) for error in self._get_errors(result)]
        self.add_error({'errors':errors,
                            'packet':entities})
        return successful_entities
        

    def _get_errored_entity_index(self,message):
        if "Error updating entity for record number" in message:
            matched = re.findall(r'\[(.*)\]', message)
            if matched:
                return int(matched[0])-1
        return None


    def _get_failed_entity(self,error,entities):
        errorKey = self._get_errored_entity_index(error.Message)
        if errorKey:
            entity = entities[errorKey]
            self.failed_entities.append(entity)
            return [entity]
        self.failed_entities.extend(entities)
        return entities


    def _get_error(self,error,entities):
        failed_entities = self._get_failed_entity(error, entities)
        return {'message':error.Message,
                'failed_entities':failed_entities}
            

class ResponseQuery(Response):
    def add_result(self,result,query):
        if result.ReturnCode == 1:
            return self.add_entities(get_result_entities(result))
        else:
            self._add_errors(self._get_errors(result),query)
        logger.debug('Adding successful result number:%s to query response',
                        len(self.response_count)
                     )
        return []
    
    def _add_errors(self,errors,query):
        for error in errors:
            self.add_error(error.Message)
        self.add_error(query.get_query_xml())


class Cursor(object):
    def __init__(self,generator):
        self._generator = generator
        
    
    def __iter__(self):
        'Returns itself as an iterator object'
        return self._generator
    
    
    def __getattr__(self,attr):
        return getattr(self._generator,attr)
    

class QueryCursor(Cursor):
    def fetch_all(self):
        return list(self._generator)    
    
    
    def fetch_one(self):
        return next(self._generator,None)
    
    
class ActionCursor(QueryCursor):
    def execute(self):
        for _ in self._generator:
            pass
    
        
class Wrapper(connection.Connection):
    outbound_entity_functions = [datetime_to_api_timezone_entity, 
                                 trim_empty_strings_entity]
    inbound_entity_functions = [datetime_to_local_timezone_entity,
                                trim_single_space_strings_entity]
    
    
    @property
    def picklist(self):
        try:
            picklist = self._picklist
        except MethodNotFound:
            picklist = Picklists(self)
            self._picklist = picklist
        return picklist
        
        
    def new(self,entity_type,**kwargs):
        entity = self.client.factory.create(entity_type)
        for k,v in iteritems(kwargs):
            setattr(entity,k,v)
        return entity


    def get_field_info(self,entity_type):
        return get_field_info(self, entity_type)


    def get_udf_info(self,entity_type):
        fields = self.client.factory.create('getUDFInfo')
        fields.psTable = entity_type
        return self.client.service.getUDFInfo(fields)

    
    def create(self,entities,**kwargs):
        return ActionCursor(self.process(entities,'create',**kwargs))


    def update(self,entities,**kwargs):
        return ActionCursor(self.process(entities,'update',**kwargs))

    
    def delete(self,entities,**kwargs):
        return ActionCursor(self.process(entities,'delete',**kwargs))    


    def process(self,entities,action,**kwargs):
        if isinstance(entities, sudsobject.Object):
            entities = [entities]
        if not entities:
            raise Exception('process entities called without anything in the entity list')
        if action not in ('create','update','delete'):
            raise Exception('action not in create update delete: {}'.format(action))
        response = ResponseAction(self)
        packet_entities = []
        packet_limit = self._get_packet_limit(**kwargs)
        for entity in entities:
            self._process_outbound_entity(entity)
            if len(packet_entities) != packet_limit:
                packet_entities.append(entity)
                continue
            packet = self._get_entity_packet(packet_entities)
            for returned_entity in self._send_packet(action, packet, response):
                yield returned_entity
            packet_entities = [entity]
        if packet_entities:
            packet = self._get_entity_packet(packet_entities)
            for returned_entity in self._send_packet(action, packet, response):
                yield returned_entity
        response.raise_or_return_entities()


    def process_inbound_entities(self,entities):
        for entity in entities:
            self._process_inbound_entity(entity)
        

    def query(self,query):
        response = ResponseQuery(self)
        return QueryCursor(self._query(query,response))


    def queries(self,queries):
        response = ResponseQuery(self)
        return QueryCursor(self._queries(queries, response))
        

    def _process_outbound_entity(self,entity):
        if not WRAPPER_DISABLE_CLEAN_ENTITIES:
            clean_entity(entity)
        process_entity(entity, self.outbound_entity_functions)


    def _process_inbound_entity(self,entity):
        process_entity(entity, self.inbound_entity_functions)


    def _query(self,query,response,**kwargs):
        finished = False
        while not finished:
            try:
                xml = query.get_query_xml()
            except AttributeError:
                xml = query            
            try:
                logger.debug('fetching query results %s/? for %s',
                             len(response.response_count) + 1,
                             query.entity_type)
                result = self.client.service.query(xml)
                logger.debug('fetched query results')
            except Exception as e:
                raise AutotaskProcessException(e,response)
            else:
                for entity in response.add_result(result, query):
                    yield entity
            if query_requires_another_call(result, query):
                highest_id = get_highest_id(result,query.minimum_id_field)
                query.set_minimum_id(highest_id)
            else:
                finished = True
        if not kwargs.get('queries',False):
            response.raise_or_return_entities()
    
    
    def _queries(self,queries,response):
        for query in queries:
            for entity in self._query(query, response, queries=True):
                    yield entity
        response.raise_or_return_entities()
    

    def _get_packet_limit(self,**kwargs):
        multiupdate = kwargs.get('bulksend',None)
        if multiupdate == False:
            return 1
        else:
            return kwargs.get('packet_limit',
                              AUTOTASK_API_ENTITY_SEND_LIMIT)
        
    
    def _get_entity_packet(self,entities):
        packet = self.client.factory.create('ArrayOfEntity')
        packet.Entity = entities
        return packet
        

    def _send_packet(self,action,packet,response):
        logger.debug('%s packet containing %s %s',
                     action,
                     len(packet.Entity),
                     get_entity_type(packet.Entity[0]))
        try:
            result = getattr(self.client.service,action)(packet)
        except Exception as e:
            # @todo the failed packet needs to come in here
            # to be available in the response exception
            logger.exception('An unhandled exception in the wrapper.')
            raise AutotaskProcessException(e,response)
        else:
            logger.debug('yielding packet response')
            return response.add_result(result, packet)

    


from pytz import timezone
USE_REQUEST_TRANSPORT_TYPE = True
REQUEST_TRANSPORT_TRANSIENT_ERROR_RETRIES = 5
REQUEST_TRANSPORT_TIMEOUT_CONNECT_WAIT = 4
REQUEST_TRANSPORT_TIMEOUT_RESPONSE_WAIT = 27
DISABLE_SSL_WARNINGS = True
SUPPORTED_SUDS_VERSION = 0.7
LOCAL_TZ='Europe/London'
AUTOTASK_API_TZ = 'America/New_York'
LOCAL_TIMEZONE = timezone(LOCAL_TZ)

MONKEY_PATCHING_ENABLED = True

AUTOTASK_API_TIMEZONE = timezone(AUTOTASK_API_TZ)
AUTOTASK_API_BASE_URL = 'https://webservices.autotask.net/atservices/1.5/atws.wsdl'
AUTOTASK_API_QUERY_RESULT_LIMIT = 500
AUTOTASK_API_QUERY_ID_LIMIT = 200
AUTOTASK_API_QUERY_DATEFORMAT = "%Y-%m-%d %H:%M:%S"
AUTOTASK_API_ENTITY_SEND_LIMIT = 200

AUTOTASK_QUERY_XML_SPACING = 3

WRAPPER_DISABLE_CLEAN_ENTITIES = False
WRAPPER_BULK_UPDATE_ENABLED = True
WRAPPER_DEFAULT_GET_ALL_ENTITIES = True

SUPPORT_FILES_ENABLED = False
SUPPORT_FILES_LOCATION = '/tmp'


#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
    'click>=6.0',
    'requests',
    'pytz',
    'suds-jurko',
    'future',
    'cached_property'
]

test_requirements = [
    # TODO: put package test requirements here
]

setup(
    name='atws',
    version='0.4.8',
    description="python-atws is a wrapper for the AutoTask SOAP webservices API",
    long_description=readme + '\n\n' + history,
    author="Matt Parr",
    author_email='matt@parr.geek.nz',
    url='https://github.com/MattParr/python-atws',
    packages=[
        'atws',
        'atws.monkeypatch'
    ],
    package_dir={'atws':
                 'atws'},
    entry_points={
        'console_scripts': [
            'create_picklist_module=atws.create_picklist_module:main'
        ]
    },
    include_package_data=True,
    install_requires=requirements,
    license="MIT license",
    zip_safe=False,
    keywords='atws Autotask API SOAP',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
    test_suite='tests',
    tests_require=test_requirements
)



'''
Created on 27 Sep 2015

@author: matt
'''
from __future__ import absolute_import
from builtins import range
import logging
import suds.client
import suds.transport as transport
from .constants import (REQUEST_TRANSPORT_TRANSIENT_ERROR_RETRIES,
                        REQUEST_TRANSPORT_TIMEOUT_CONNECT_WAIT,
                        REQUEST_TRANSPORT_TIMEOUT_RESPONSE_WAIT,
                        AUTOTASK_API_BASE_URL,
                        DISABLE_SSL_WARNINGS,
                        USE_REQUEST_TRANSPORT_TYPE)
                        
from requests.exceptions import ConnectTimeout, Timeout, ReadTimeout, SSLError

logger = logging.getLogger(__name__)
suds_version = float(suds.__version__[:3])
logger.debug('Suds Version %s', suds_version)

if suds_version < 0.7:
    logger.warning('Suds version (%s) not ideal. Please use 0.7',
                   suds_version)

import requests
import functools
import traceback
try:
    from cStringIO import StringIO 
except ImportError:
    try:
        from StringIO import StringIO
    except ImportError:
        from io import BytesIO as StringIO

MAX_RETRIES = REQUEST_TRANSPORT_TRANSIENT_ERROR_RETRIES

def get_zone_info(username):
    client = suds.client.Client(url=AUTOTASK_API_BASE_URL)
    return client.service.getZoneInfo(username)


def get_zone_url(username):
    zone_info = get_zone_info(username)
    return zone_info.URL


def get_zone_id(username):
    zone_info = get_zone_info(username)
    return zone_info.CI


def get_zone_wsdl(username):
    url = get_zone_url(username)
    try:
        return url.replace('asmx','wsdl')
    except AttributeError:
        logging.exception('failed to resolve zone')
        raise ValueError('Username:{} failed to resolve to a zone'.format(
            username))


def get_connection_url(**kwargs):
    try:
        url = kwargs['url']
        assert url
    except (KeyError, AssertionError):
        url = get_zone_wsdl(kwargs['username'])
    return url

    
def disable_warnings():
    import requests.packages
    requests.packages.urllib3.disable_warnings()
    
    
def connect(**kwargs):
    client_options = kwargs.setdefault('client_options',{})
    if DISABLE_SSL_WARNINGS:
        disable_warnings()
    if USE_REQUEST_TRANSPORT_TYPE:
        session = requests.Session()
        session.auth = (kwargs['username'],kwargs['password'])
        session.mount("http://", 
                      requests.adapters.HTTPAdapter(max_retries=MAX_RETRIES))
        session.mount("https://", 
                      requests.adapters.HTTPAdapter(max_retries=MAX_RETRIES))
        transport = client_options.setdefault('transport',
                                              RequestsTransport(session))
    
    url = get_connection_url(**kwargs)
    client_options['url'] = url
    obj = kwargs.get('atws_version',Connection)
    return obj(**kwargs)


def handle_errors(f):
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except requests.HTTPError as e:
            raise transport.TransportError(
                'Error in requests\n' + traceback.format_exc(),
                e.response.status_code,
            )
        except requests.RequestException:
            raise transport.TransportError(
                'Error in requests\n' + traceback.format_exc(),
                000
            )
    return wrapper


class RequestsTransport(transport.Transport):
    # @todo: handle transient errors by retrying
    # ideally, we should only have to raise on a few errors
    # like destination unreachable, but not timeouts or transport
    # failure eg: SSLError below
    def __init__(self, session=None):
        transport.Transport.__init__(self)
        self._session = session or requests.Session()
        self.timeout = (REQUEST_TRANSPORT_TIMEOUT_CONNECT_WAIT,
                        REQUEST_TRANSPORT_TIMEOUT_RESPONSE_WAIT)
        
        
    @handle_errors
    def open(self, request):
        for attempt in range(1,REQUEST_TRANSPORT_TRANSIENT_ERROR_RETRIES + 1):
            try:
                resp = self._session.get(request.url,timeout=self.timeout)
                break
            except (SSLError, ConnectTimeout, Timeout, ReadTimeout):
                if attempt == REQUEST_TRANSPORT_TRANSIENT_ERROR_RETRIES:
                    raise
                continue
        return StringIO(resp.content)
    
    
    @handle_errors
    def send(self, request):
        for attempt in range(1,REQUEST_TRANSPORT_TRANSIENT_ERROR_RETRIES + 1):
            try:
                resp = self._session.post(
                                          request.url,
                                          data=request.message,
                                          headers=request.headers,
                                          )
                break
            except (SSLError, Timeout):
                if attempt == REQUEST_TRANSPORT_TRANSIENT_ERROR_RETRIES:
                    raise
                continue
        return transport.Reply(
            resp.status_code,
            resp.headers,
            resp.content,
        )


class Connection(object):
    '''
    classdocs
    '''
    def __init__(self, **kwargs):
        '''
        Constructor
        '''
        self.kwargs = kwargs
        try:
            self.client = kwargs['client']
        except KeyError:
            options = kwargs.get('client_options',{})
            self.client = suds.client.Client(**options)
    
    
    def __getattr__(self,attr):
        return getattr(self.client.service,attr)
    
    